//! Utilities for formatting of a table
pub use crate::{
    cell::{Align, Justify, Padding, PaddingBuilder},
    table::{Border, BorderBuilder, HorizontalLine, Separator, SeparatorBuilder, VerticalLine},
};
